#include "Vcf2Map.h"

#define physicalDistPerCM 1000000

Vcf2Map::Vcf2Map(){
 Init();
}

Vcf2Map::Vcf2Map(int wins){
 Init();
 winSize = wins;
 geno.resize(winSize);
 varData.resize(winSize);
}

void Vcf2Map::Init()
{
 winSize = 0;;
 varCounter = 0;
 GT_idx = -1;
 DP_idx = -1;
 minDP = 0.0;
 maxDP = -1.0;
 gt2int["0/0"] = 0;
 gt2int["0/1"] = 1;
 gt2int["0/1"] = 2;
}

Vcf2Map::~Vcf2Map(){}

void Vcf2Map::SetMaxR2(double maxr2){maxR2 = maxr2; }
void Vcf2Map::SetWinSize(int winsize ){winSize = winsize; geno.resize(winSize); varData.resize(winSize);}
void Vcf2Map::SetMinMAF(double maf ){minMAF = maf;}
void Vcf2Map::SetMinAvgDP(double dp ){minDP = dp;}
void Vcf2Map::SetMaxAvgDP(double dp ){maxDP = dp;}
void Vcf2Map::SetMaxMissingRate(double r){maxMissingRate = r;}
void Vcf2Map::SetOutMapFile(FILE* out) { outMap = out; fprintf(outMap, "Chromosome\tMarker\tMap(cM)\tPosition(bp)\n"); }

void Vcf2Map::LoadIncludeList(IFILE incFH){
  if(incFH==NULL) return;

  String s;
  StringArray tokens;
  int lines=0;

  printf("Loading included variant list (");

   while(!ifeof(incFH))
    {
      s.ReadLine(incFH);
      tokens.ReplaceTokens(s);
      if(tokens.Length()==0) continue;

      lines++;
      if(lines==1) { printf("skipping the header: %s) ...\n", s.c_str()); continue; } //skip the header

      if(tokens.Length()<2) { fprintf(stderr, "Less than 4 columns are in the map file!\n"); exit(1); }

      std::string chr = std::string(tokens[0].c_str());
      if(chr.substr(0,3).compare("chr")==0) chr = chr.substr(3, chr.size());
      int pos = tokens[1].AsInteger();

      cleanVar[std::pair<std::string, int>(chr, pos)].chr = chr;
      cleanVar[std::pair<std::string, int>(chr, pos)].chr = pos;

      if(tokens.Length()>2) cleanVar[std::pair<std::string, int>(chr, pos)].EUR_AF = tokens[2].AsDouble();
      if(tokens.Length()>3) cleanVar[std::pair<std::string, int>(chr, pos)].AFR_AF = tokens[3].AsDouble();
      if(tokens.Length()>4) cleanVar[std::pair<std::string, int>(chr, pos)].ASN_AF = tokens[4].AsDouble();
      if(tokens.Length()>5) cleanVar[std::pair<std::string, int>(chr, pos)].AF = tokens[5].AsDouble();

    }
}

void Vcf2Map::LoadMapFile(IFILE mapFH){
  if(mapFH==NULL) return;

  String s;
  StringArray tokens;
  int lines=0;

  printf("Loading genetic map file (");

   while(!ifeof(mapFH))
    {
      s.ReadLine(mapFH);
      tokens.ReplaceTokens(s);
      if(tokens.Length()==0) continue;

      lines++;
      if(lines==1) { printf("skipping the header: %s) ...\n", s.c_str()); continue; } //skip the header

      if(tokens.Length()<4) { fprintf(stderr, "Less than 4 columns are in the map file!\n"); exit(1); }

      std::string chr = std::string(tokens[0].c_str());
      if(chr.substr(0,3).compare("chr")==0) chr = chr.substr(3, chr.size());
      int pos = tokens[1].AsInteger();
      double dist = tokens[3].AsDouble();

      gmap[chr].chr.push_back(chr);
      gmap[chr].position.push_back(pos);
      gmap[chr].distance.push_back(dist);
    }
}


//prune LD and generate genetic map of pruned variants
//three filtering criteria used: 1000G clean dataset, MAF and missing rates. More may be added
void Vcf2Map::ProcessVCF(std::string vcf)
{
 VCFInputFile vin(vcf.c_str());
 std::string pre_chr = "";

 printf("Pruning markers in LD and generating pruned map file...\n");

 while(vin.readRecord())
  {
      VCFRecord &r = vin.getVCFRecord();
      VCFPeople &people = r.getPeople();

      std::string chr = r.getChrom();
      if(chr.substr(0,3).compare("chr")==0) chr = chr.substr(3, chr.size());
      int pos = r.getPos();

      if(cleanVar.size()>0 && cleanVar.find(std::pair<std::string, int>(chr, pos))==cleanVar.end()) continue; // remove all variants that are not in the 1000G cleaned dataset

      if(GT_idx<0)
      {
        GT_idx = r.getFormatIndex("GT");
        if(GT_idx<0) {
          fprintf(stderr, "NO GT field was found. Please check the vcf file at chr:%s and position:%d", chr.c_str(), pos);
          exit(1);
        }
        DP_idx = r.getFormatIndex("DP");
      }

      VCFIndividual* indv;
      bool missing = false;

      int non_missing_count = 0;
      int missing_count = 0;
      int AC = 0;
      int total_dp = 0;
      int dp = 0;

      for(int i=0; i<people.size(); i++) //this loop is to get AF estimated by counting AC
      {
        indv = people[i];
        if(pids.size()>0 && pids[(*indv).getName()]!=1) continue; // skip non-founders only when ped is provided

	std::string gt = (*indv).get(GT_idx, &missing).toStr();

	if(missing) { missing_count++; continue; }

	if(DP_idx>=0) dp = (*indv).get(DP_idx, &missing).toInt();
	total_dp += dp;

	if(gt.compare("0/1")==0) AC++;
	if(gt.compare("1/1")==0) AC+=2;
	non_missing_count++;
      }

      double missing_rate = double(missing_count)/double(missing_count+non_missing_count);

      if(missing_rate>maxMissingRate) continue; //remove variants with high missing rates

      double AF = AC==0 ? 0.0 : double(AC)/double(non_missing_count*2);
      if(AF>0.5) AF = 1-AF;
      if(AF<minMAF) continue;  //remove variants with low MAF

      double avg_dp = double(total_dp)/double(missing_count+non_missing_count);

      if(avg_dp<minDP || avg_dp>maxDP && maxDP>0) continue; // remove low or high depth variants

      double gt_val;

      for(int i=0; i<people.size(); i++) //this loop will store genotype and imputed mean of missing genotype
      {
        indv = people[i];
        if(pids.size()>0 && pids[(*indv).getName()]!=1) continue; // skip non-founders only when ped is provided

        std::string gt = (*indv).get(GT_idx, &missing).toStr();

	if(missing) gt_val = 2*AF*(1-AF);
	else gt_val = gt2int[gt];
	geno[varCounter].push_back(gt_val); 
      }

      varData[varCounter].chr = chr;
      varData[varCounter].position = pos;

      varCounter++;

      if(varCounter>=winSize || chr.compare(pre_chr)!=0 && pre_chr.size()>0) {
	//PrintGenoMatrix(); printf("-------------\n");
	PruneLD();
	chr.compare(pre_chr)!=0 ? Reset_keepone() : Reset();
      }
      pre_chr = chr;
  }
  PruneLD();
}

void Vcf2Map::PruneLD()
{
 double r2;

 for(int i=0; i<varCounter; i++)
 {
   if(varData[i].chr.compare(varData[0].chr)!=0) break;
   if(removed.find(i)!=removed.end()) continue;
   for(int j=i+1; j<varCounter; j++)
   {
     r2 = CalcPairwiseR2(geno[i], geno[j]);
     if(r2>maxR2) removed[j]=1;
   }
 }

 GeneratePrunedMap();
}

void Vcf2Map::GeneratePrunedMap()
{
 int offset;
 double genet_dist, dist_per_bp;
 std::string chr = varData[0].chr;

 for(int i=0; i<varCounter; i++)
 {
  if(varData[i].chr.compare(varData[0].chr)!=0) break;

  if(removed.find(i)!=removed.end()) continue;

  if(gmap.find(chr)==gmap.end()) {
   genet_dist = double(varData[i].position)/physicalDistPerCM;
   fprintf(outMap, "%s\t%s:%d\t%f\t%d\n", varData[i].chr.c_str(), varData[i].chr.c_str(), varData[i].position, genet_dist, varData[i].position);
   continue;
  }


 std::vector<int>& position = gmap[chr].position;
 std::vector<double>& distance = gmap[chr].distance;
 int& counter = gmap[chr].posCounter;

  while(1)
  {
   if(counter==distance.size()-1)
   {
     offset = varData[i].position - position[counter];
     genet_dist = distance[counter] + double(offset)/physicalDistPerCM;
     break;
   }
   else if (varData[i].position >= position[counter] && varData[i].position <= position[counter+1])
   {
     dist_per_bp = (distance[counter+1] - distance[counter]) / double(position[counter+1] - position[counter]);
     offset = varData[i].position - position[counter];
     genet_dist = distance[counter] + double(offset) * dist_per_bp;

     break;
   }
   counter++;
  }
  fprintf(outMap, "%s\t%s:%d\t%f\t%d\n", varData[i].chr.c_str(), varData[i].chr.c_str(), varData[i].position, genet_dist, varData[i].position);
 }
}

void Vcf2Map::Reset(){
 for(int i=0; i<varCounter; i++)
	geno[i].clear();
 removed.clear();
 varCounter = 0;
}

void Vcf2Map::Reset_keepone(){
 if(varCounter==0) return;
 geno[0] = geno[varCounter-1];
 varData[0].chr = varData[varCounter-1].chr;
 varData[0].position = varData[varCounter-1].position;

 for(int i=1; i<varCounter; i++)
   geno[i].clear();
 removed.clear();
 varCounter = 1;
}

void Vcf2Map::GetFounders(Pedigree& ped)
{
  pids.clear();
  for(int i=0; i<ped.familyCount; i++)
    for(int j=0; j<ped.families[i]->count; j++)
      {
        Person *p;
        p = ped.persons[ped.families[i]->path[j]];
        if(p->isFounder()) pids[std::string(p->pid.c_str())]=1;
      }
}

double Vcf2Map::CalcPairwiseR2(std::vector<double>& x, std::vector<double>& y)
{
 double r2  = 0.0;

 int n = x.size();
 if(n==0) return(r2);

 double sum_x = 0.0;
 double sum_y = 0.0;
 double sum_x2 = 0.0;
 double sum_y2 = 0.0;
 double sum_xy = 0.0;

 for(int i=0; i<n; i++)
 {
  sum_x += x[i];
  sum_y += y[i];
  sum_x2 += x[i]*x[i];
  sum_y2 += y[i]*y[i];
  sum_xy += x[i]*y[i];
 }

 r2 = pow((n*sum_xy-sum_x*sum_y),2)/((n*sum_x2-sum_x*sum_x) * (n*sum_y2-sum_y*sum_y));
 return(r2);
}

void Vcf2Map::PrintGenoMatrix()
{
 for(int i=0; i<varCounter; i++)
 {
  for(int j=0; j<geno[i].size(); j++)
    printf("%.2f ", geno[i][j]);
  printf("\n");
 }
}
